from __future__ import annotations
import os.path as osp
from os import PathLike
import pathlib
from pathlib import (
  PurePath,
  PureWindowsPath,
  PurePosixPath )
import re

from .pattern import (
  PatternError,
  tr_glob,
  tr_path,
  tr_rel_join )

#===============================================================================
class PathMatcher:
  r"""Pattern matching similar to '.gitignore'

  Parameters
  ----------
  pattern:
    See notes below on pattern formatting.
  negate:
    A match to this pattern negates an existing match of the same name.
  dironly:
    This pattern is to only match the name of a directory.
  relative:
    This pattern is to match relative paths instead of just the base name.

  Notes
  -----
  * https://git-scm.com/docs/gitignore#_pattern_format
  * An optional prefix "!" which negates the pattern; any matching file excluded
    by a previous pattern will become included again. It is not possible to
    re-include a file if a parent directory of that file is excluded.
    Git doesn't list excluded directories for performance reasons, so any
    patterns on contained files have no effect, no matter where they are defined.
    Put a backslash ("\") in front of the first "!" for patterns that begin with
    a literal "!", for example, "\!important!.txt".
  * The slash / is used as the directory separator. Separators may occur at the
    beginning, middle or end of the .gitignore search pattern.
  * If there is a separator at the beginning or middle (or both) of the pattern,
    then the pattern is relative to the directory level of the particular
    .gitignore file itself. Otherwise the pattern may also match at any level
    below the .gitignore level.
  * If there is a separator at the end of the pattern then the pattern will only
    match directories, otherwise the pattern can match both files and directories.
  * For example, a pattern doc/frotz/ matches doc/frotz directory, but not
    a/doc/frotz directory; however frotz/ matches frotz and a/frotz that is a
    directory (all paths are relative from the .gitignore file).
  * An asterisk "*" matches anything except a slash. The character "?" matches
    any one character except "/".
  * The range notation, e.g. [a-zA-Z], can be used
    to match one of the characters in a range. See fnmatch(3) and the FNM_PATHNAME
    flag for a more detailed description.
    This logic rests on the idea that a character class
    cannot be an empty set. e.g. [] would not match anything, so is not allowed.
    This means that []] is valid since the the first pair "[]" cannot close
    a valid set.
    Likewise, [!] cannot complement an empty set, since this would be equivalent
    to *, meaning it should instead match "!".
    [!] -> match "!"
    [!!] -> match any character that is not "!"
    []] -> match "]"
    [!]] -> match any character that is not "]"
    []!] -> match "]" or "!"
  * Two consecutive asterisks ("**") in patterns matched against full pathname
    may have special meaning:
  * A leading "**" followed by a slash means match in all directories. For example,
    "**/foo" matches file or directory "foo" anywhere, the same as pattern "foo".
    "**/foo/bar" matches file or directory "bar" anywhere that is directly under
    directory "foo".
  * A trailing "/**" matches everything inside. For example, "abc/**" matches all
    files inside directory "abc", relative to the location of the .gitignore file,
    with infinite depth.
  * A slash followed by two consecutive asterisks then a slash matches zero or
    more directories. For example, "a/**/b" matches "a/b", "a/x/b", "a/x/y/b"
    and so on.
  * The meta-characters "*", "?", and "[" may be escaped by backslash.

  """
  #-----------------------------------------------------------------------------
  def __init__(self,
    pattern: str,
    negate: bool = False,
    dironly: bool = False,
    relative: bool = False):

    pattern = str(pattern).strip()
    _pattern = pattern

    if pattern.startswith('!'):
      # An optional prefix "!" which negates the pattern
      negate = True
      pattern = pattern[1:]

    elif pattern.startswith(r'\!'):
      # Put a backslash ("\") in front of the first "!" for patterns that begin
      # with a literal "!", for example, "\!important!.txt".
      pattern = pattern[1:]

    if pattern.endswith('/'):
      # If there is a separator at the end of the pattern then the pattern will
      # only match directories
      dironly = True
      pattern = pattern[:-1]

    if pattern.count('/') > 0:
      # If there is a separator at the beginning or middle (or both) of the
      # pattern, then the pattern is relative to the directory level of the
      # particular .gitignore file itself.
      relative = True

      if pattern.startswith('/'):
        pattern = pattern[1:]

      elif pattern.startswith('./'):
        pattern = pattern[2:]

    self._pattern = _pattern
    self._pattern_tr, self._pattern_segs = tr_glob(pattern)
    self._rec = re.compile( self._pattern_tr )
    self._match = self._rec.match
    self.negate = negate
    self.dironly = dironly
    self.relative = relative

    #DEBUG print(f"{self._pattern} -> {self._pattern_tr}")
    #DEBUG print('  ' + '\n  '.join([str(seg) for seg in self._pattern_segs]))

  #-----------------------------------------------------------------------------
  def __str__(self):
    return self._pattern

  #-----------------------------------------------------------------------------
  def __repr__(self):
    args = [f"'{self._pattern}'"]

    for attr in ['negate', 'dironly', 'relative']:
      if getattr(self, attr):
        args.append(f'{attr} = True')

    args = ', '.join(args)

    return f"{type(self).__name__}({args})"

  #-----------------------------------------------------------------------------
  def match(self, path: PathLike) -> bool:
    """
    Parameters
    ----------
    path:

    Returns
    -------
    matched :
      True if the ``path`` matches this pattern

    """

    if not isinstance(path, PurePath):
      path = PurePath(path)

    _path = tr_path(path)
    #DEBUG print('match', path, '->', _path)
    return self._match(_path)

  #-----------------------------------------------------------------------------
  __call__ = match

  #-----------------------------------------------------------------------------
  def nt(self, path: PathLike) -> bool:
    """Convenience method to force match as a Windows path
    """
    return self(PureWindowsPath(path))

  #-----------------------------------------------------------------------------
  def posix(self, path: PathLike) -> bool:
    """Convenience method to force match as a POSIX path
    """
    return self(PurePosixPath(path))

#===============================================================================
class PathFilter:
  """A combination of file patters applied relative to a given 'start' directory

  Parameters
  ----------
  patterns:
  start:
  """
  #-----------------------------------------------------------------------------
  def __init__(self,
      patterns: list[str|PathMatcher] = None,
      start: PathLike|None = None):

    if patterns is None:
      patterns = []

    self.patterns = [
      p if isinstance(p, PathMatcher) else PathMatcher(p)
      for p in patterns ]

    _start = None

    if start is not None:
      if not isinstance(start, PurePath):
        start = PurePath(start)

      _start = tr_path(start)
      #DEBUG print('start', start, '->', _start)

    self.start = start
    self._start = _start

  #-----------------------------------------------------------------------------
  def filter(self,
      dir: PathLike,
      fnames: list[str],
      dnames: list[str]|None = None,
      feasible: set[str]|None = None) -> set[str]:
    """Filter a list of names in a directory

    Parameters
    ----------
    dir:
      Directory containing ``dnames`` and ``fnames``.
    fnames:
      List of file (non-directory) names in ``dir``.
    dnames:
      List of directory names in ``dir``.

      .. note::

        If None, any fnames ending with '/' will be used as (directory) dnames.

    feasible:
      The current feasible set of names (from either dnames or fnames) that have
      been matched.

    Returns
    -------
    feasible:
      Updated feasible set of matched names. It is possible that the input
      feasible set contains names that are *not* in the output if a pattern
      negates an existing match.
    """

    if not isinstance(dir, PurePath):
      dir = PurePath(dir)

    _dir = tr_path(dir)
    #DEBUG print('dir', dir, '->', _dir)

    if dnames is None:
      dnames, fnames = partition(lambda x: x.endswith(osp.sep), fnames)

    fnames = [d.rstrip(osp.sep) for d in fnames]
    dnames = [d.rstrip(osp.sep) for d in dnames]

    return self._filter(_dir, fnames, dnames, feasible)

  #-----------------------------------------------------------------------------
  def _filter(self, dir, fnames, dnames, feasible = None):
    """Internal method, assumes dir has already been converted to posix, and
    fnames/dnames must be separatly given.
    """
    dname_paths = tr_rel_join(self._start, dir, dnames)
    fname_paths = tr_rel_join(self._start, dir, fnames)
    name_paths = dname_paths + fname_paths

    if feasible is None:
      feasible = set()

    #DEBUG print(f"  {self.start}, {dir}")
    #DEBUG print(f"    fnames: {fname_paths}")
    #DEBUG print(f"    dnames: {dname_paths}")

    # Can only match directories, filter out other names
    for pattern in self.patterns:
      op = feasible.difference if pattern.negate else feasible.union
      _name_paths = dname_paths if pattern.dironly else name_paths
      match = pattern._match

      if pattern.relative:
        feasible = op({ name for name, path in _name_paths if match(path) })
      else:
        feasible = op({ name for name, path in _name_paths if match(name) })

      #DEBUG print(f"    - {repr(pattern)} -> {feasible}")

    #DEBUG print(f"    {feasible}")

    return feasible


#===============================================================================
def contains(a, b):
  a = str(a)
  b = str(b)
  return a == osp.commonpath([a, b])

#===============================================================================
def partition(test, vals):
  """Separates a single list into two lists

  Parameters
  ----------
  test : callable
  vals : list

  Returns
  -------
  x, y: (list, list)
    The first list contains elememts of ``vals`` where ``test`` returned true.
    The second list contains all other elements.
  """
  x = list()
  y = list()

  for val in vals:
    if test(val):
      x.append(val)
    else:
      y.append(val)

  return x, y

#===============================================================================
def partition_dir(dir, names):
  """Separates a list of names into those that are directorys and all others.
  """
  return partition(
    lambda name: not osp.isdir(osp.join(dir, name)),
    names )

#===============================================================================
def combine_ignore_patterns(*patterns):
  """Creates a callable as ``ignore``

  Parameters
  ----------
  *patterns : PathMatcher

  Returns
  -------
  callable(dir, names) -> matches
  """
  # TODO: implement as callable object instead of closure for possible inspection

  def _ignore_patterns(dir, names):
    dir = PurePath(dir)

    #DEBUG print(f"dir: {dir}")

    feasible = set()

    fnames, dnames = partition_dir(dir, names)

    #DEBUG print(f"  dnames: {dnames}")
    #DEBUG print(f"  fnames: {fnames}")

    _dir = tr_path(dir)

    for pattern in patterns:
      feasible = pattern._filter(_dir, fnames, dnames, feasible)

    return feasible

  return _ignore_patterns



from __future__ import annotations
import sys
import re

#===============================================================================
def _gen_nonprintable():
  test = ''

  ns = [ [0,], ]

  for i in range(1, sys.maxunicode+1):
    c = chr(i)
    test += c

    if not ( c.isprintable() or c in '\n\t' ):
      n = ns[-1]

      if i == n[-1] + 1:
        if len(n) == 1:
          n.append(i)
        else:
          n[-1] = i
      else:
        ns.append([i,])

  # print(len(test), test.isprintable())
  # print(len(ns))
  # print(ns)

  return ns, test

#===============================================================================
def gen_nonprintable():
  """Method used to generate a regex for matchiing all non-printable unicode
  characters, except for newlines '\\n' and tabs '\\t'.
  """
  ns, test = _gen_nonprintable()

  def fmt(i):
    if i < 2**8:
      return rf'\x{i:02X}'
    elif i < 2**16:
      return rf'\u{i:04X}'
    else:
      return rf'\U{i:08X}'

  # format character ranges as unicode literals
  ns = [
    fmt(n[0]) if len(n) == 1 else ( fmt(n[0]) + '-' + fmt(n[-1]) )
    for n in ns ]

  nonprintable = "  r'["
  _nonprintable = '['

  line_max = 75
  line_len = len(nonprintable)

  for n in ns:
    if line_len + len(n) > line_max:
      nonprintable += "'\n  r'"
      line_len = 0

    nonprintable += n
    _nonprintable += n
    line_len += len(n)

  nonprintable += "]'"
  _nonprintable += ']'

  _nonprintable = re.compile(
    _nonprintable.encode('utf-8').decode('unicode_escape'),
    re.UNICODE )

  test = _nonprintable.sub('', test)
  test = re.sub(r'[\t\n]', '', test)

  # print( len(test), test.isprintable() )

  assert( test.isprintable() )

  return nonprintable

#===============================================================================
if __name__ == '__main__':
  print( gen_nonprintable() )



from __future__ import annotations
import os
import shutil

#===============================================================================
def meson_option_arg(k, v):
  """Convert python key-value pair to meson ``-Dkey=value`` option
  """
  if isinstance(v, bool):
    v = ({True: 'true', False: 'false'})[v]

  return f'-D{k}={v}'

#===============================================================================
def meson(
  pyproj,
  logger,
  options,
  work_dir,
  src_dir,
  build_dir,
  prefix,
  setup_args,
  compile_args,
  install_args,
  build_clean,
  runner):
  """Run meson setup, compile, install commands

  Parameters
  ----------
  pyproj : :class:`PyProjBase <partis.pyproj.pyproj.PyProjBase>`
  logger : logging.Logger
  options : dict
  src_dir : pathlib.Path
  build_dir : pathlib.Path
  prefix : pathlib.Path
  setup_args : list[str]
  compile_args : list[str]
  install_args : list[str]
  build_clean : bool
  """

  if not shutil.which('meson'):
    raise ValueError("The 'meson' program not found.")

  if not shutil.which('ninja'):
    raise ValueError("The 'ninja' program not found.")

  os.environ['MESON_FORCE_BACKTRACE'] = '1'

  # TODO: ensure any paths in setup_args are normalized
  if not build_clean:
    # skip setup if the build directory already populated
    setup_args = list()
  else:
    # only run setup if the build directory does not already exist (or is empty)
    setup_args = [
      'meson',
      'setup',
      *setup_args,
      '--prefix',
      str(prefix),
      *[ meson_option_arg(k,v) for k,v in options.items() ],
      str(build_dir),
      str(src_dir) ]

  compile_args = [
    'meson',
    'compile',
    *compile_args,
    '-C',
    str(build_dir) ]

  install_args = [
    'meson',
    'install',
    *install_args,
    '--no-rebuild',
    '-C',
    str(build_dir) ]


  if setup_args:
    runner.run(setup_args)

  runner.run(compile_args)
  runner.run(install_args)



from __future__ import annotations
import re
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from pathlib import (
  Path,
  PurePath,
  PurePosixPath)

from collections.abc import (
  Mapping,
  Sequence,
  Iterable )

from .validate import (
  OPTIONAL,
  OPTIONAL_NONE,
  REQUIRED,
  valid,
  union,
  restrict,
  valid_dict,
  valid_list,
  ValidationError,
  as_list )

from .norms import (
  marker_evaluated,
  scalar,
  scalar_list,
  empty_str,
  nonempty_str,
  str_list,
  nonempty_str_list,
  norm_bool,
  norm_path,
  norm_path_to_os )

from .pep import (
  CompatibilityTags,
  purelib_compat_tags,
  norm_printable,
  valid_dist_name,
  norm_dist_version,
  norm_dist_author_dict,
  norm_dist_extra,
  norm_entry_point_group,
  norm_entry_point_name,
  norm_entry_point_ref,
  norm_dist_keyword,
  norm_dist_classifier,
  norm_dist_url )

#===============================================================================
class dynamic(valid_list):
  value_valid = restrict(
    'version',
    'description',
    'readme',
    'authors',
    'maintainers',
    'license',
    'dynamic',
    'requires-python',
    'dependencies',
    'optional-dependencies',
    'keywords',
    'classifiers',
    'urls',
    'scripts',
    'gui-scripts',
    'entry-points' )

#===============================================================================
class readme(valid_dict):
  # a string at top-level interpreted as a path to the readme file
  proxy_key = 'file'
  allow_keys = list()
  min_keys = [
    ('file', 'text')]
  mutex_keys = [
    ('file', 'text')]
  default = {
    # NOTE: file paths should initially be given as a POSIX path,
    # but converted to current OS path so it may be read.
    'file': valid(OPTIONAL, PurePosixPath, Path),
    'text': valid(OPTIONAL, nonempty_str, norm_printable) }

#===============================================================================
class license(valid_dict):
  allow_keys = list()
  min_keys = [
    ('file', 'text')]
  default = {
    'file': valid(OPTIONAL, PurePosixPath, Path),
    'text': valid(OPTIONAL, nonempty_str, norm_printable) }

#===============================================================================
class author(valid_dict):
  validator = valid(norm_dist_author_dict)
  allow_keys = list()
  min_keys = [
    ('name', 'email')]
  default = {
    'name': valid(str),
    'email': valid(str) }

#===============================================================================
class authors(valid_list):
  value_valid = valid(author)

#===============================================================================
class maintainer(author):
  pass

#===============================================================================
class maintainers(valid_list):
  value_valid = valid(maintainer)

#===============================================================================
class dependencies(valid_list):
  value_valid = valid(norm_printable, Requirement, str)

#===============================================================================
class optional_dependency_group(dependencies):
  pass

#===============================================================================
class optional_dependencies(valid_dict):
  key_valid = valid(norm_dist_extra)
  value_valid = valid(optional_dependency_group)

#===============================================================================
class entry_point_group(valid_dict):
  key_valid = valid(norm_entry_point_name)
  value_valid = valid(norm_entry_point_ref)

#===============================================================================
class scripts(entry_point_group):
  pass

#===============================================================================
class gui_scripts(entry_point_group):
  pass

#===============================================================================
class entry_points(valid_dict):
  key_valid = valid(norm_entry_point_group)
  value_valid = valid(entry_point_group)

  # PEP 621
  # > Build back-ends MUST raise an error if the metadata defines a
  # > [project.entry-points.console_scripts] or [project.entry-points.gui_scripts]
  # > table, as they would be ambiguous in the face of [project.scripts]
  # > and [project.gui-scripts], respectively.
  forbid_keys = [
    'scripts',
    'console_scripts',
    'gui-scripts',
    'gui_scripts' ]

#===============================================================================
class keywords(valid_list):
  value_valid = valid(norm_dist_keyword)

#===============================================================================
class classifiers(valid_list):
  value_valid = valid(norm_dist_classifier)

#===============================================================================
def norm_dist_url_item(kv):
  return norm_dist_url(*kv)

#===============================================================================
class urls(valid_dict):
  item_valid = valid(norm_dist_url_item)

#===============================================================================
class project(valid_dict):
  allow_keys = list()
  require_keys = [
    'name']
  default = {
    'dynamic': dynamic,
    'name': valid(valid_dist_name),
    'version': valid('0.0.0', norm_dist_version),
    'description': valid(str, norm_printable),
    # must be optional because there is no default value
    'readme': valid(OPTIONAL, readme),
    # must be optional because there is no default value
    'license': valid(OPTIONAL, license),
    'authors': valid(authors),
    'maintainers': valid(maintainers),
    'keywords': valid(keywords),
    'classifiers': valid(classifiers),
    'urls': valid(urls),
    'requires-python': valid(str, norm_printable, SpecifierSet, str),
    'dependencies': valid(dependencies),
    'optional-dependencies': valid(optional_dependencies),
    'scripts': valid(scripts),
    'gui-scripts': valid(gui_scripts),
    'entry-points': valid(entry_points) }

#===============================================================================
class build_requires(dependencies):
  pass

#===============================================================================
class path_parts(valid_list):
  value_valid = valid(nonempty_str)

#===============================================================================
class build_system(valid_dict):
  allow_keys = list()
  require_keys = [
    'build-backend']
  default = {
    'requires': build_requires,
    'build-backend': norm_entry_point_ref,
    'backend-path': valid(OPTIONAL_NONE, path_parts) }

#===============================================================================
def compat_tag(v):
  return CompatibilityTags(*v)

#===============================================================================
class compat_tags(valid_list):
  _as_list = valid(as_list)
  _min_len = 1
  value_valid = valid(compat_tag)

#===============================================================================
class pyproj_prep(valid_dict):
  allow_keys = list()
  require_keys = [
    'entry' ]
  default = {
    'entry': norm_entry_point_ref,
    'kwargs': dict }

#===============================================================================
class pyproj_dist_prep(pyproj_prep):
  pass

#===============================================================================
class pyproj_dist_source_prep(pyproj_prep):
  pass

#===============================================================================
class pyproj_dist_binary_prep(pyproj_prep):
  pass

#===============================================================================
class pyproj_build_target(valid_dict):
  allow_keys = list()
  require_keys = [
    'entry' ]
  deprecate_keys = [
    # name changed to be more general
    ('compile', 'enabled')]
  default = {
    'enabled': valid(True, marker_evaluated),
    # NOTE: default builder from backward compatibility
    'entry': valid('partis.pyproj.builder:meson', norm_entry_point_ref),
    'options': dict,
    'env': dict,
    # NOTE: paths should start as POSIX, but transformed to current OS
    'work_dir': valid('.', PurePosixPath, Path),
    'src_dir': valid('.', PurePosixPath, Path),
    'build_dir': valid('build/tmp', PurePosixPath, Path),
    'prefix': valid('build', PurePosixPath, Path),
    'setup_args': nonempty_str_list,
    'compile_args': nonempty_str_list,
    'install_args': nonempty_str_list,
    'build_clean': valid(True, norm_bool) }

#===============================================================================
class pyproj_meson(valid_dict):
  """

  .. deprecated:: 0.1.0
    Replaced by more general :class:`pyproj_build_target`

  """
  allow_keys = list()
  default = {
    'compile': valid(False, norm_bool),
    'src_dir': valid('.', nonempty_str, norm_path, norm_path_to_os),
    'build_dir': valid('build/meson', nonempty_str, norm_path, norm_path_to_os),
    'prefix': valid('build', nonempty_str, norm_path, norm_path_to_os),
    'setup_args': nonempty_str_list,
    'compile_args': nonempty_str_list,
    'install_args': nonempty_str_list,
    'options': dict,
    'build_clean': valid(True, norm_bool) }

#===============================================================================
class pyproj_targets(valid_list):
  _as_list = valid(as_list)
  value_valid = valid(pyproj_build_target)

#===============================================================================
class ignore_list(valid_list):
  _as_list = valid(as_list)
  value_valid = valid(nonempty_str)

#===============================================================================
class glob_list(valid_list):
  _as_list = valid(as_list)
  value_valid = valid(nonempty_str)

#===============================================================================
class include(valid_dict):
  allow_keys = list()
  # a string at top-level interpreted as 'match'
  proxy_key = 'glob'
  # TODO: how to normalize patterns?
  default = {
    'glob': valid(nonempty_str),
    'rematch': valid(r'.*', nonempty_str, re.compile),
    'replace': valid('{0}', nonempty_str)}

#===============================================================================
class include_list(valid_list):
  _as_list = valid(as_list)
  value_valid = valid(include)

#===============================================================================
class pyproj_dist_copy(valid_dict):
  # a string at top-level interpreted as 'src'
  proxy_key = 'src'
  # take 'dst' from 'src' if not set
  proxy_keys = [('dst', 'src')]
  deprecate_keys = [('glob', 'include')]
  allow_keys = list()
  min_keys = [
    ('src', 'glob') ]
  default = {
    # NOTE: file paths should initially be given as a POSIX path,
    # but converted to current OS path so it may be read.
    'src': valid(REQUIRED, PurePosixPath, Path),
    # the destination path in the archive should remain as a POSIX path
    'dst': valid(REQUIRED, PurePosixPath),
    'include': include_list,
    'ignore': ignore_list }

#===============================================================================
class pyproj_dist_copy_list(valid_list):
  value_valid = valid(pyproj_dist_copy)

#===============================================================================
class pyproj_dist_scheme(valid_dict):
  allow_keys = list()
  default = {
    'ignore': ignore_list,
    'copy': pyproj_dist_copy_list }

#===============================================================================
class pyproj_dist_binary(valid_dict):
  allow_keys = list()
  default = {
    'build_number': valid(OPTIONAL_NONE, int),
    'build_suffix': valid(OPTIONAL_NONE, str),
    'compat_tags': valid(purelib_compat_tags(), compat_tags),
    'prep': valid(OPTIONAL_NONE, pyproj_dist_binary_prep),
    'ignore': ignore_list,
    'copy': pyproj_dist_copy_list,
    'data': pyproj_dist_scheme,
    'headers': pyproj_dist_scheme,
    'scripts': pyproj_dist_scheme,
    'purelib': pyproj_dist_scheme,
    'platlib': pyproj_dist_scheme }

#===============================================================================
class pyproj_dist_source(valid_dict):
  allow_keys = list()
  default = {
    'prep': valid(OPTIONAL, pyproj_dist_source_prep),
    'ignore': ignore_list,
    'copy': pyproj_dist_copy_list,
    'add_legacy_setup': valid(False, norm_bool) }

#===============================================================================
class pyproj_dist(valid_dict):
  allow_keys = list()
  default = {
    'prep': valid(OPTIONAL, pyproj_dist_prep),
    'ignore': ignore_list,
    'source': pyproj_dist_source,
    'binary': pyproj_dist_binary }

#===============================================================================
class pyproj_config(valid_dict):
  key_valid = valid(norm_dist_extra)
  value_valid = union(scalar, scalar_list)

#===============================================================================
class pyproj(valid_dict):
  allow_keys = list()
  default = {
    'config': pyproj_config,
    'prep': valid(OPTIONAL, pyproj_prep),
    'dist': pyproj_dist,
    'targets': pyproj_targets }
  deprecate_keys = [('meson', 'targets')]

#===============================================================================
class tool(valid_dict):
  require_keys = ['pyproj']
  default = {
    'pyproj': pyproj }

#===============================================================================
class pptoml(valid_dict):
  allow_keys = list()
  require_keys = [
    'project',
    'build-system']
  default = {
    'project': valid(REQUIRED, project),
    'build-system': valid(REQUIRED, build_system),
    'tool': valid(OPTIONAL, tool) }



from __future__ import annotations
import os
import io
import csv
from pathlib import (
  PurePosixPath)
from ..norms import (
  hash_sha256,
  email_encode_items )
from ..pep import (
  norm_dist_build,
  norm_dist_compat,
  compress_dist_compat,
  norm_dist_filename )
from ..pkginfo import PkgInfo
from .dist_zip import dist_zip
from ..path import (
  subdir,
  PathError )

#===============================================================================
def pkg_name(dir):
  if dir.endswith('.py'):
    return dir[:-3]

  return dir

#===============================================================================
class dist_binary_wheel( dist_zip ):
  """Build a binary distribution :pep:`427`, :pep:`491` wheel file ``*.whl``

  Parameters
  ----------
  pkg_info : :class:`PkgInfo <partis.pyproj.pkginfo.PkgInfo>`
  build : str
    Build tag. Must start with a digit, or be an empty string.
  compat : List[ Tuple[str,str,str] ] | List[ :class:`CompatibilityTags <partis.pyproj.norms.CompatibilityTags>` ]
    List of build compatability tuples of the form ( py_tag, abi_tag, plat_tag ).
    e.g. ( 'py3', 'abi3', 'linux_x86_64' )
  outdir : str
    Path to directory where the wheel file should be copied after completing build.
  tmpdir : None | str
    If not None, uses the given directory to place the temporary wheel file before
    copying to final location.
    My be the same as outdir.
  logger : None | :class:`logging.Logger`
    Logger to use.
  gen_name : str
    Name to use as the 'Generator' of the wheel file

  Example
  -------

  .. code-block:: python

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:

      import os
      import os.path

      pkg_dir = os.path.join( tmpdir, 'src', 'my_package' )
      out_dir = os.path.join( tmpdir, 'build' )

      os.makedirs( pkg_dir )

      with open( os.path.join( pkg_dir, 'module.py' ), 'w' ) as fp:
        fp.write("print('hello')")

      from partis.pyproj import (
        PkgInfo,
        dist_binary_wheel )

      pkg_info = PkgInfo(
        project = dict(
          name = 'my-package',
          version = '1.0' ) )


      with dist_binary_wheel(
        pkg_info = pkg_info,
        outdir = out_dir ) as bdist:

        bdist.copytree(
          src = pkg_dir,
          dst = 'my_package' )

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0427
  * https://www.python.org/dev/peps/pep-0491
  * https://www.python.org/dev/peps/pep-0660

  """
  #-----------------------------------------------------------------------------
  def __init__( self, *,
    pkg_info,
    build = '',
    compat = None,
    outdir = None,
    tmpdir = None,
    logger = None,
    gen_name = None ):

    if not compat:
      compat = [ ( 'py3', 'none', 'any' ), ]

    if not isinstance( pkg_info, PkgInfo ):
      raise ValueError(f"pkg_info must be instance of PkgInfo: {pkg_info}")

    self.pkg_info = pkg_info

    if gen_name is None:
      gen_name = f'{type(self).__module__}.{type(self).__name__}'

    self.build = norm_dist_build( build )

    self.compat = [
      norm_dist_compat( py_tag, abi_tag, plat_tag )
      for py_tag, abi_tag, plat_tag in compat ]

    self.top_level = list()
    self.purelib = True
    self.gen_name = str(gen_name)

    wheel_name_parts = [
      self.pkg_info.name_normed,
      self.pkg_info.version,
      self.build,
      *compress_dist_compat( self.compat ) ]

    wheel_name_parts = [
      norm_dist_filename(p)
      for p in wheel_name_parts
      if p != '' ]

    self.base_path = PurePosixPath('-'.join( wheel_name_parts[:2] ))
    self.base_tag = '-'.join( wheel_name_parts[-3:] )

    self.dist_info_path = PurePosixPath(str(self.base_path) + '.dist-info')
    self.data_path = PurePosixPath(str(self.base_path) + '.data')
    self.metadata_path = self.dist_info_path.joinpath('METADATA')
    self.entry_points_path = self.dist_info_path.joinpath('entry_points.txt')
    self.wheel_path = self.dist_info_path.joinpath('WHEEL')
    self.record_path = self.dist_info_path.joinpath('RECORD')

    self.data_paths = [
      'data',
      'headers',
      'scripts',
      'purelib',
      'platlib' ]

    super().__init__(
      outname = '-'.join( wheel_name_parts ) + '.whl',
      outdir = outdir,
      tmpdir = tmpdir,
      logger = logger,
      named_dirs = {
        'dist_info' : self.dist_info_path,
        **{ k : self.data_path.joinpath(k) for k in self.data_paths } } )

  #-----------------------------------------------------------------------------
  def finalize( self ):

    if self.record_hash:
      return self.record_hash

    self.check_top_level()

    self.write(
      dst = self.metadata_path,
      data = self.pkg_info.encode_pkg_info() )

    if self.pkg_info.license_file:
      self.write(
        dst = self.dist_info_path.joinpath(self.pkg_info.license_file),
        data = self.pkg_info.license_file_content )

    self.write(
      dst = self.dist_info_path.joinpath('top_level.txt'),
      data = '\n'.join( self.top_level ) )

    self.write(
      dst = self.entry_points_path,
      data = self.pkg_info.encode_entry_points() )

    self.write(
      dst = self.wheel_path,
      data = self.encode_dist_info_wheel() )

    record_data, self.record_hash = self.encode_dist_info_record()

    self.write(
      dst = self.record_path,
      data = record_data,
      # NOTE: the record itself is not recorded in the record
      record = False )

    return self.record_hash

  #-----------------------------------------------------------------------------
  def check_top_level( self ):
    """Discover the package top_level from record entries
    """

    top_level = set()

    purelib = self.named_dirs['purelib']

    platlib = self.named_dirs['platlib']

    for file, (hash, size) in self.records.items():
      # check files added to purelib and platlib.
      try:
        top_level.add(pkg_name(subdir(purelib, file).parts[0]))
        continue
      except PathError:
        pass

      try:
        top_level.add(pkg_name(subdir(platlib, file).parts[0]))
        self.purelib = False
        continue
      except PathError:
        pass

      try:
        subdir(self.dist_info_path, file)
        subdir(self.data_path, file)
      except PathError:
        # check any other files that aren't in .dist-info or .data
        top_level.add(pkg_name(file.parts[0]))

    self.top_level = [ dir for dir in top_level if dir ]

  #-----------------------------------------------------------------------------
  def encode_dist_info_wheel( self ):
    """Generate content for .dist_info/WHEEL

    Returns
    -------
    content : bytes
    """

    headers = [
      ( 'Wheel-Version', '1.0' ),
      ( 'Generator', self.gen_name ),
      ( 'Root-Is-Purelib', str(self.purelib).lower() ),
      *[ ( 'Tag', '-'.join( compat ) ) for compat in self.compat ],
      ( 'Build', self.build ) ]

    return email_encode_items( headers = headers )

  #-----------------------------------------------------------------------------
  def encode_dist_info_record( self ):
    """Generate content for .dist_info/RECORD

    Returns
    -------
    content : bytes
    hash : str
      sha256 hash of the record file data
    """

    record = io.StringIO()
    record_csv = csv.writer(record)

    # the record file itself is listed in records, but the hash of the record
    # file cannot be included in the file.
    _records = {**self.records, self.record_path: ('', '')}

    for file, (hash, size) in _records.items():
      hash = f'sha256={hash}' if hash else ''
      record_csv.writerow([os.fspath(file), hash, size])

    content = record.getvalue().encode('utf-8')

    hash, size = hash_sha256(content)

    return content, hash



from __future__ import annotations
import re
from copy import copy
from pathlib import Path
from collections.abc import (
  Sequence,
  Mapping)
# from string import Template

from .validate import (
  ValidationError,
  FileOutsideRootError)
from .path import (
  subdir,
  resolve)

namespace_sep = re.compile(r"[\.\[\]]")

_idpattern = re.compile(r"""(?:
  # references a template variable
  [A-Z_][A-Z0-9_]* # Python identifier
  (?:
    \.[A-Z_][A-Z0-9_]* # attribute access by Python identifier
    |
    \[-?[0-9]+\])* # integer index (potentially negative)
  |
  \/ # forward slash separate for building path
  |
  (?<=\/)\.\. # double-dot (parent directory), must follow a slash
  |
  '[A-Z0-9\-_\.]+' # string literal, quotes to be removed
  )+
  """,
  re.IGNORECASE|re.VERBOSE)

_group_pattern = re.compile(
  # NOTE: handles escaped '$' and missing closing brace
  r"\$(?:(?P<escaped>\$)|{ *(?P<braced>[^\s\}]+) *(?:}|(?P<unterminated>$|[^}])))",
  re.IGNORECASE)

#===============================================================================
class TemplateError(ValidationError):
  ...

#===============================================================================
class NamespaceError(ValidationError):
  ...

#===============================================================================
class Template:
  r"""Template support nested mappings and paths using :class:`Namespace`
  """

  #-----------------------------------------------------------------------------
  def __init__(self, template):
    self.template = template

  #-----------------------------------------------------------------------------
  def substitute(self, namespace: Mapping = None, /, **kwargs):
    if namespace is None:
      namespace = kwargs

    elif kwargs:
      raise TypeError("Cannot use both namespace and kwargs")

    if not isinstance(namespace, Namespace):
      namespace = Namespace(namespace)

    def _handler(m):
      if m.group('escaped'):
        return '$'

      if m.group('unterminated') is not None:
        raise TemplateError(f"Unterminated template substitution: {m.group()}")

      name = m.group('braced').strip()

      if not _idpattern.fullmatch(name):
        raise TemplateError(f"Invalid template substitution: {name}")

      return str(namespace[name])

    return _group_pattern.sub(_handler, self.template)

#===============================================================================
class Namespace(Mapping):
  r"""Namespace (potentially nested) mapping for using with :class:`Template`

  Parameters
  ----------
  data:
    Mapping for names to values. Note that changes to the namespace will also
    change the data. Making a shallow copy of the namespace also make a shallow
    copy of the data.
  root:
    If given, absolute path to project root, used to resolve relative paths and ensure
    any derived paths are within this parent directory.
  """
  __slots__ = ['data', 'root']

  #-----------------------------------------------------------------------------
  def __init__(self, data: Mapping, *, root: Path = None):
    self.data = data
    self.root = root

  #-----------------------------------------------------------------------------
  def __iter__(self):
    return iter(self.data)

  #-----------------------------------------------------------------------------
  def __len__(self):
    return len(self.data)

  #-----------------------------------------------------------------------------
  def __setitem__(self, name, value):
    self.data[name] = value

  #-----------------------------------------------------------------------------
  def __getitem__(self, key):
    raw_segments = key.split('/')
    segments = []

    for name in raw_segments:
      if len(name) == 0 or name == '..':
        # empty segment
        segments.append(name)

      elif name.startswith("'"):
        # string literal, remove quotes
        segments.append(name[1:-1])

      else:
        # variable name lookup
        segments.append(self.lookup(name))

    if len(segments) == 1:
      return segments[0]

    if self.root is None:
      out = Path(*segments)

    else:
      root = self.root
      out = type(root)(*segments)

      if not out.is_absolute():
        out = root/out

      if isinstance(root, Path):
        # NOTE: ignored if root is a pure path
        out = resolve(out)

      if not subdir(root, out, check = False):
        raise FileOutsideRootError(
          f"Must be within project root directory:"
          f"\n  file = \"{out}\"\n  root = \"{root}\"")

    return out

  #-----------------------------------------------------------------------------
  def __copy__(self):
    cls = type(self)
    obj = cls.__new__(cls)
    obj.data = copy(self.data)
    obj.root = self.root
    return obj

  #-----------------------------------------------------------------------------
  def lookup(self, name):
    parts = namespace_sep.split(name)
    data = self.data

    try:
      cur = []

      for k in parts:
        if k:
          if isinstance(data, Mapping):
            data = data[k]
          elif not isinstance(data, (str,bytes)) and isinstance(data, Sequence):
            i = int(k)
            data = data[i]
          else:
            raise NamespaceError(f"Expected mapping or sequence for '{k}': {type(data).__name__}")

          cur.append(k)

    except (KeyError,TypeError,IndexError) as e:
      raise NamespaceError(f"Invalid key '{k}' of name '{name}': {str(e)}") from None

    return data

#===============================================================================
def template_substitute(
    value: bool|int|str|Path|Mapping|Sequence,
    namespace: Mapping):
  r"""Recursively performs template substitution based on type of value
  """

  if not isinstance(namespace, Namespace):
    namespace = Namespace(namespace)

  if isinstance(value, (bool,int)):
    # just handles case where definitely not a template
    return value

  cls = type(value)

  if isinstance(value, str):
    return cls(Template(value).substitute(namespace))

  if isinstance(value, Path):
    return cls(*(
      Template(v).substitute(namespace)
      for v in value.parts))

  if isinstance(value, Mapping):
    return cls({
      k: template_substitute(v, namespace)
      for k,v in value.items()})

  if isinstance(value, Sequence):
    return cls([
      template_substitute(v, namespace)
      for v in value])


  raise TypeError(f"Unknown template value type: {value}")




from __future__ import annotations
import os
import os.path as osp
import sys
import shutil
import logging
from logging import (
  basicConfig,
  getLogger,
  Logger)
import tempfile
import re

from pathlib import (
  Path,
  PurePath,
  PurePosixPath)

from collections.abc import (
  Mapping,
  Sequence )

from . import (
  valid_keys,
  mapget,
  dist_build,
  PkgInfoReq,
  PyProjBase,
  dist_binary_wheel,
  dist_source_targz )

#===============================================================================
def backend_init(
  root: str|Path = '',
  config_settings: dict|None = None,
  logger: Logger|None = None ):
  """Called to inialialize the backend upon a call to one of the hooks

  Parameters
  ----------
  root :
    Directory containing 'pyproject.toml'
  logger :
    Logger to use

  Returns
  -------
  PyProjBase
  """

  # NOTE: this is mainly used for debugging, since front-ends don't seem to have
  # an option to set logging level for the backend.
  root_logger = getLogger()

  if not root_logger.handlers:
    basicConfig(
      level = logging.INFO,
      format = "{message}",
      style = "{" )

  root = Path(root)
  logger = logger or getLogger( __name__ )

  pyproj = PyProjBase(
    root = root,
    config_settings = config_settings,
    logger = logger )

  return pyproj


#-----------------------------------------------------------------------------
def get_requires_for_build_sdist(
  config_settings: dict|None = None ):
  """
  Note
  ----
  This hook MUST return an additional list of strings containing PEP 508
  dependency specifications, above and beyond those specified in the
  pyproject.toml file. These dependencies will be installed when calling the
  build_sdist hook.

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-sdist
  """

  return list()

#-----------------------------------------------------------------------------
def build_sdist(
  dist_directory,
  config_settings: dict|None = None ):
  """
  Note
  ----
  Must build a .tar.gz source distribution and place it in the specified
  dist_directory. It must return the basename (not the full path) of the
  .tar.gz file it creates, as a unicode string.

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0517/#build-sdist
  """

  pyproj = backend_init(config_settings = config_settings)

  pyproj.dist_prep()

  pyproj.dist_source_prep()

  with dist_source_targz(
    pkg_info = pyproj.pkg_info,
    outdir = dist_directory,
    logger = pyproj.logger ) as dist:

    pyproj.dist_source_copy(
      dist = dist )

  return dist.outname

#-----------------------------------------------------------------------------
def get_requires_for_build_wheel(
  config_settings: dict|None = None ):
  """
  Note
  ----
  This hook MUST return an additional list of strings containing
  PEP 508 dependency specifications, above and beyond those specified in the
  pyproject.toml file, to be installed when calling the build_wheel or
  prepare_metadata_for_build_wheel hooks.

  Note
  ----
  pip appears to not process environment markers for deps returned
  by get_requires_for_build_*, and may falsly report
  ``ERROR: Some build dependencies...conflict with the backend dependencies...``

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-wheel
  """

  pyproj = backend_init(config_settings = config_settings)

  # filter out any dependencies already listed in the 'build-system'.
  # NOTE: pip appears to not process environment markers for deps returned
  # by get_requires_for_build_*, and may falsly report
  # > ERROR: Some build dependencies...conflict with the backend dependencies...
  build_requires = pyproj.build_requires - set([
    PkgInfoReq(r)
    for r in mapget( pyproj.pptoml, 'build-system.requires', list() ) ])

  reqs = [ str(r) for r in build_requires ]

  pyproj.logger.info(f'get_requires_for_build_wheel: {reqs}')

  return reqs

#-----------------------------------------------------------------------------
def prepare_metadata_for_build_wheel(
  metadata_directory,
  config_settings: dict|None = None ):
  """
  Note
  ----
  Must create a .dist-info directory containing wheel metadata inside the
  specified metadata_directory (i.e., creates a directory like
  {metadata_directory}/{package}-{version}.dist-info/).

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0517/#prepare-metadata-for-build-wheel
  """

  pyproj = backend_init(config_settings = config_settings)

  # TODO: abstract 'wheel metadata' from needing to actually make a dummy wheel file
  with dist_binary_wheel(
    pkg_info = pyproj.pkg_info,
    outdir = metadata_directory,
    logger = pyproj.logger ) as dist:

    pass


  import zipfile
  with zipfile.ZipFile( dist.outpath ) as fp:
    fp.extractall(metadata_directory)

  # NOTE: dist_info_path is a POSIX path, need to convert to OS path first
  # PIP assums the return value is a string
  return os.fspath(Path(dist.dist_info_path))

#-----------------------------------------------------------------------------
def build_wheel(
  wheel_directory,
  config_settings: dict|None = None,
  metadata_directory = None ):
  """
  Note
  ----
  Must build a .whl file, and place it in the specified wheel_directory.
  It must return the basename (not the full path) of the .whl file it creates,
  as a unicode string.


  See Also
  --------
  * https://www.python.org/dev/peps/pep-0517/#build-wheel
  """

  pyproj = backend_init(config_settings = config_settings)

  pyproj.dist_prep()

  pyproj.dist_binary_prep()

  with dist_binary_wheel(
    pkg_info = pyproj.pkg_info,
    build = dist_build(
      pyproj.binary.get('build_number', None),
      pyproj.binary.get('build_suffix', None) ),
    compat = pyproj.binary.compat_tags,
    outdir = wheel_directory,
    logger = pyproj.logger ) as dist:

    pyproj.dist_binary_copy(
      dist = dist )

  pyproj.logger.info(
    f"Top level packages {dist.top_level}")

  return dist.outname

#-----------------------------------------------------------------------------
# def prepare_metadata_for_build_editable(
#   metadata_directory,
#   config_settings = None ):
#   pass


#-----------------------------------------------------------------------------
# def build_editable(
#   wheel_directory,
#   config_settings = None,
#   metadata_directory = None ):
#   pass

#===============================================================================
class UnsupportedOperation( Exception ):
  """
  Note
  ----
  If the backend cannot produce an dist because a dependency is missing,
  or for another well understood reason, it should raise an exception of a
  specific type which it makes available as UnsupportedOperation on the
  backend object.

  See Also
  --------
  * https://www.python.org/dev/peps/pep-0517/
  """
  pass


from __future__ import annotations
import sys
from os import (
  curdir,
  pardir,
  fspath)
from os.path import (
  realpath)
from pathlib import (
  Path,
  PurePath)

#===============================================================================
class PathError(ValueError):
  pass

#===============================================================================
def resolve(path: Path):
  r"""Backport of latest Path.resolve behavior
  """
  return type(path)(realpath(fspath(path)))

#===============================================================================
def _concretize(comps: list[str]) -> list[str]|None:
  r"""Mostly equivalent to :func:`os.path.normpath`, except for the cases where
  a concrete path is not possible or would be truncated.

  For example, the path `a/../b` can be normalized to the concrete path `b`,
  but `a/../../b` depends the name of a's parent directory.
  """

  new_comps = []

  for comp in comps:
    if not comp or comp == curdir:
      continue

    if comp == pardir:
      if not new_comps:
        # concrete path not possible
        return None

      new_comps.pop()
    else:
      new_comps.append(comp)

  return new_comps

#===============================================================================
def _subdir(_start: list[str], _path: list[str]) -> list[str]|None:
  r"""Concrete path relative to start, or `None` if path is not a sub-directory
  """

  if (_start := _concretize(_start)) is None:
    return None

  if (_path := _concretize(_path)) is None:
    return None

  n = len(_start)

  if len(_path) < n or _path[:n] != _start:
    return None

  return _path[n:]

#===============================================================================
def subdir(start: PurePath, path: PurePath, check: bool = True) -> PurePath|None:
  """Relative path, restricted to sub-directories.

  Parameters
  ----------
  start:
    Starting directory.
  path:
    Directory to compute relative path to, *must* be a sub-directory of `start`.
  check:
    If True, raises exception if not a subdirectory. Otherwise returns None.

  Returns
  -------
  rpath:
    Relative path from `start` to `path`.
  """

  _rpath = _subdir(start.parts, path.parts)

  if _rpath is None:
    if check:
      raise PathError(f"Not a subdirectory of {start}: {path}")

    return None

  return type(path)(*_rpath)



from __future__ import annotations
import os
import os.path as osp
from pathlib import PurePath
import sys
import shutil
import logging
import tempfile


#===============================================================================
def legacy_setup_content( pyproj, sdist ):
  """

  Note
  ----
  Assumes that `pyproj.dist_source_prep` has already been run, and all files have
  already been added to `sdist`
  """

  # NOTE: the build requirements are added to the installation requirements,
  # since legacy installation does not truly support the 'setup_requires' method
  # provided by setuptools

  # NOTE: 'wheel' is added as a build dependency because pip will not even try
  # to run 'bdist_wheel' if 'wheel' isn't already installed

  build_requires = pyproj.build_requires | set(['wheel', ])

  pkg_info = pyproj.pkg_info.add_dependencies(
    deps = build_requires )

  requires = '\n'.join([
    str(d) for d in  pkg_info.requires_dist ]).encode('utf-8')

  sources = '\n'.join( [
    os.fspath(file)
    for file, (hash, size) in sdist.records.items() ] ).encode('utf-8')

  with open( osp.join( osp.dirname(__file__), '_legacy_setup.py' ), 'r' ) as fp:
    legacy_setup_py = fp.read()

  build_requires = [ str(r) for r in build_requires ]

  legacy_setup_py = legacy_setup_py.format(
    egg_info_name = f"'{pyproj.pkg_info.name_normed}.egg-info'",
    pkg_info = pkg_info.encode_pkg_info(),
    requires = requires,
    build_requires = str(build_requires),
    sources = sources,
    top_level = b'',
    entry_points = pkg_info.encode_entry_points(),
    build_backend = f"'{pyproj.build_backend}'",
    backend_path = pyproj.backend_path )

  sdist.write(
    dst = f'{sdist.base_path}/setup.py',
    data = legacy_setup_py )



from __future__ import annotations
import os

#===============================================================================
def tail(path, n, bufsize = 1024, encoding = 'utf-8'):
  """Reads the last n lines from a file

  Parameters
  ----------
  path : str
  n : int
    Max number of lines to read from the end of the file
  bufsize : int
    Number of bytes to buffer at a time.

  Returns
  -------
  lines : List[str]
    Up to ``n`` lines from the end of the file
  """

  bufsize = int(bufsize)
  bufsize = max(1, bufsize)

  n = int(n)
  n = max( 0, n )

  buf = bytes()
  nlines = 0

  head = 0

  with open( path, 'rb' ) as fp:
    # total number of bytes in the file
    tot = fp.seek( 0, os.SEEK_END )

    head = tot

    while nlines < n and head > 0:
      # NOTE: the number of newline characters is one less than number of 'lines'
      nread = min( head, bufsize )
      head -= nread

      fp.seek( head, os.SEEK_SET )

      _buf = fp.read( nread )

      # NOTE: this is not exact, since pairs of '\r\n' may happen on the read
      # boundary, but should count approximately the number of LF, CR, and CR+LF.
      # LF -> 1 + 0 + 0 = 1
      # CR -> 0 + 1 + 0 = 1
      # CR+LF -> 1 + 1 - 1 = 1
      nlines += _buf.count(b'\n') + _buf.count(b'\r') - _buf.count(b'\r\n')

      buf = _buf + buf

  if nlines > 0 and head > 0:
    # remove everything before first newline to ensure only complete lines are kept
    i = buf.index(b'\n')
    buf = buf[(i+1):]

  res = buf.decode(encoding, errors = 'replace')
  lines = res.split('\n')[-n:]

  return lines



from __future__ import annotations
import os
from collections.abc import Callable
from pathlib import (
  Path,
  PurePosixPath)
from logging import (
  getLogger,
  Logger)
import re
from abc import(
  ABC,
  abstractmethod )
from ..norms import (
  hash_sha256 )
from ..validate import (
  validating )

#===============================================================================
class dist_base( ABC ):
  """Builder for file distribution

  Parameters
  ----------
  outname:
    Name of output file.
  outdir:
    Path to directory where the file should be copied after completing build.
  tmpdir:
    If not None, uses the given directory to place the temporary file(s) before
    copying to final location.
    May be the same as outdir.
  named_dirs:
    Mapping of specially named directories within the distribution.
    By default a named directory { 'root' : '.' } will be added,
    unless overridden with another directory name. Must be in form of Posix path.
  logger:
    Logger to which any status information is to be logged.

  Attributes
  ----------
  outpath : Path
    Path to final output file location
  named_dirs : dict[str, PurePosixPath]
    Mapping of specially named directories within the distribution
  opened : bool
    Build temporary file has been opened for writing
  finalized : bool
    Build temporary file has been finalized.
  closed : bool
    Build temporary file has been closed
  copied : bool
    Build temporary has been copied to ``outpath`` location
  records : dict[PurePosixPath, tuple[str, int]]
    Recorded mapping of path to hash, and size (bytes) of files added to distribution
  record_hash : None | str
    Final hash value of the record after being finalized

    .. note::

      Not all distribution implementations will create a hash of the record

  """
  #-----------------------------------------------------------------------------
  def __init__( self,
    outname: str,
    outdir: Path|None = None,
    tmpdir: Path|None = None,
    logger: Logger|None = None,
    named_dirs: dict[str, PurePosixPath]|None = None ):

    if logger is None:
      logger = getLogger(type(self).__name__)

    if named_dirs is None:
      named_dirs = dict()

    named_dirs = { name: PurePosixPath(dir) for name, dir in named_dirs.items() }

    self.outname = str(outname)
    self.outdir = Path(outdir) if outdir else Path.cwd()
    self.outpath = self.outdir.joinpath(self.outname)
    self.tmpdir = Path(tmpdir) if tmpdir else None
    self.logger = logger
    self.named_dirs = {
      'root' : PurePosixPath(),
      **named_dirs }

    self.records = dict()
    self.record_bytes = None
    self.record_hash = None

    self.opened = False
    self.finalized = False
    self.closed = False
    self.copied = False

  #-----------------------------------------------------------------------------
  def exists( self,
    dst ):
    """Behaviour similar to os.path.exists for entries in the distribution file

    Parameters
    ----------
    dst : str | PurePosixPath

    Returns
    -------
    exists : bool
    """
    return False

  #-----------------------------------------------------------------------------
  def write( self,
    dst: PurePosixPath,
    data: bytes,
    mode: int|None = None,
    record: bool = True ):
    """Write data into the distribution file

    Parameters
    ----------
    dst :
    data :
    mode :
    record :
      Add file to the record

    """

    if record:
      self.record(
        dst = dst,
        data = data )

  #-----------------------------------------------------------------------------
  def makedirs( self,
    dst: PurePosixPath,
    mode: int|None = None,
    exist_ok: bool = False,
    record: bool = True ):
    """Behaviour similar to os.makedirs into the distribution file

    Parameters
    ----------
    dst :
    mode :
    exist_ok :
    record :
      Add file to the record

    Note
    ----
    Some archive file types do not need to explicitly create directories, but this
    is given in case an implementation needs to create a directory before creating
    files within the directory.

    """

    if not exist_ok and self.exists( dst ):
      raise ValueError(f"Build file already has entry: {dst}")

  #-----------------------------------------------------------------------------
  def copyfile( self,
    src: Path,
    dst: PurePosixPath,
    mode: int|None = None,
    exist_ok: bool = False,
    record: bool = True ):
    """Behavior similar to shutil.copyfile into the distribution file

    Parameters
    ----------
    src :
    dst :
    mode :
    exist_ok :
    record :
      Add file to the RECORD
    """
    src = Path(src)
    dst = PurePosixPath(dst)

    if not src.exists():
      raise ValueError(f"Source file not found: {src}")

    if not exist_ok and self.exists( dst ):
      raise ValueError(f"Build file already has entry: {dst}")

    self.logger.debug( f'copyfile {src}' )

    if mode is None:
      mode = src.stat().st_mode

    with open( src, "rb" ) as fp:
      self.write(
        dst = dst,
        data = fp,
        mode = mode,
        record = record )

    return dst

  #-----------------------------------------------------------------------------
  def copytree( self,
    src: Path,
    dst: PurePosixPath,
    ignore: Callable[[Path, list[str]], list[str]]|None = None,
    exist_ok: bool = False,
    record: bool = True ):
    """Behavior similar to shutil.copytree into the distribution file

    Parameters
    ----------
    src :
    dst :
    ignore :
      If not None, ``callable(src, names) -> ignored_names``

      See :func:`shutil.copytree`
    exist_ok : bool
    record : bool
      Add all files to the RECORD
    """
    src = Path(src)
    dst = PurePosixPath(dst)

    if not src.exists():
      raise ValueError(f"Source directory not found: {src}")

    self.logger.debug( f'copytree {src}' )

    # returns an iterator of DirEntry
    entries = list(os.scandir(src))

    if ignore is not None:
      ignored_names = ignore(
        src,
        [ x.name for x in entries ] )

      if len(ignored_names) > 0:
        self.logger.debug( f'ignoring {ignored_names}' )

        entries = [ entry
          for entry in entries
          if entry.name not in ignored_names ]

    for entry in entries:
      src_path = src.joinpath(entry.name)
      dst_path = dst.joinpath(entry.name)

      mode = entry.stat().st_mode

      with validating(key = re.sub( r"[^\w\d]+", "_", entry.name )):
        if entry.is_dir():

          self.makedirs(
            dst = dst_path,
            mode = mode,
            # TODO: separate option for existing directories? like `dirs_exist_ok`
            exist_ok = True,
            record = record )

          self.copytree(
            src = src_path,
            dst = dst_path,
            ignore = ignore,
            exist_ok = exist_ok,
            record = record )

        else:

          self.copyfile(
            src = src_path,
            dst = dst_path,
            mode = mode,
            exist_ok = exist_ok,
            record = record )

    return dst

  #-----------------------------------------------------------------------------
  def open( self ):
    """Opens the temporary distribution file

    Returns
    -------
    self : :class:`dist_base <partis.pyproj.dist_base.dist_base>`
    """

    if self.opened:
      raise ValueError("distribution file has already been opened")

    self.logger.info( f'building {self.outname}' )

    try:

      self.create_distfile()

      self.opened = True

      return self

    except:
      self.close(
        finalize = False,
        copy = False )

      raise

  #-----------------------------------------------------------------------------
  def record( self,
    dst: PurePosixPath,
    data: bytes) -> tuple[str, int]:
    """Creates a record for an added file

    This produces an sha256 hash of the data and associates a record with the item

    Parameters
    ----------
    dst:
      Path of item within the distribution
    data:
      Binary data that was added

    """

    self.assert_recordable()

    hash, size = hash_sha256(data)

    dst = PurePosixPath(dst)
    record = (hash, size)

    self.logger.debug(f'record {dst}: {record}')
    self.records[dst] = record

    return record

  #-----------------------------------------------------------------------------
  def close( self,
    finalize: bool = True,
    copy: bool = True ):
    """Closes the temporary distribution file

    Parameters
    ----------
    finalize :
      If true, finalizes the temporary distribution file before closing
    copy :
      If true, copies the temporary file to final location after closing

    """

    if self.closed:
      return

    if finalize and not self.finalized:
      self.logger.info( f'finalizing {self.outname}' )
      self.finalize()
      self.finalized = True

    self.close_distfile()
    self.closed = True

    if copy and not self.copied:
      self.logger.info( f'copying {self.outname} -> {self.outdir}' )
      self.copy_distfile()
      self.copied = True

    self.remove_distfile()

  #-----------------------------------------------------------------------------
  def __enter__(self):
    """Opens the temporary distribution file upon entering the context

    Returns
    -------
    self : :class:`dist_base <partis.pyproj.dist_base.dist_base>`
    """
    if self.opened:
      return self

    return self.open()

  #-----------------------------------------------------------------------------
  def __exit__(self, type, value, traceback ):
    """Closes the temporary distribution file upon exiting the context

    If no exception occured in the context, then the temporary distribution is finalized
    and copied to it's final locations
    """

    # only finalize if there was not an error
    no_error = ( type is None )

    self.close(
      finalize = no_error,
      copy = no_error )

    # don't handle any errors here
    return None

  #-----------------------------------------------------------------------------
  def assert_open( self ):
    """Raises exception if temporary file is not open
    """
    if self.opened and not self.closed:
      return

    raise ValueError("distribution file is not open")

  #-----------------------------------------------------------------------------
  def assert_recordable( self ):
    """Raises exception if temporary file has already been finalized
    """
    self.assert_open()

    if not self.finalized:
      return

    raise ValueError("distribution record has already been finalized.")

  #-----------------------------------------------------------------------------
  @abstractmethod
  def create_distfile( self ):
    """Creates and opens a temporary distribution file into which files are copied
    """
    raise NotImplementedError('')

  #-----------------------------------------------------------------------------
  @abstractmethod
  def finalize( self ) -> str|None:
    """Finalizes the distribution file before being closed

    Returns
    -------
    record_hash :
      sha256 hash of the record

    Note
    ----
    Not all distribution implementations will create/return a hash of the record
    """
    raise NotImplementedError('')

  #-----------------------------------------------------------------------------
  @abstractmethod
  def close_distfile( self ):
    """Closes the temporary distribution file
    """
    raise NotImplementedError('')

  #-----------------------------------------------------------------------------
  @abstractmethod
  def copy_distfile( self ):
    """Copies the temporary distribution file to it's final location
    """
    raise NotImplementedError('')

  #-----------------------------------------------------------------------------
  @abstractmethod
  def remove_distfile( self ):
    """Deletes the temporary distribution file
    """
    raise NotImplementedError('')



from .utils import (
  PathError,
  subdir,
  resolve)

from .pattern import (
  PathPatternError,
  PatternError )

from .match import (
  PathMatcher,
  PathFilter,
  contains,
  partition,
  partition_dir,
  combine_ignore_patterns )



#===============================================================================
from .dist_base import dist_base
from .dist_zip import dist_zip
from .dist_targz import dist_targz

from .dist_source import (
  dist_source_targz,
  dist_source_dummy )

from .dist_binary import (
  dist_binary_wheel )

from .dist_copy import (
  FileOutsideRootError,
  dist_iter,
  dist_copy )




#===============================================================================
from .validate import (
  ValidationError,
  ValidationWarning,
  validating,
  valid_type,
  valid_keys,
  mapget,
  as_list )

from .norms import (
  marker_evaluated,
  scalar,
  scalar_list,
  empty_str,
  nonempty_str,
  str_list,
  nonempty_str_list,
  norm_bool,
  norm_path,
  norm_path_to_os,
  norm_mode,
  norm_zip_external_attr,
  norm_data,
  b64_nopad,
  hash_sha256,
  email_encode_items,
  TimeEncode )

from .pep import (
  CompatibilityTags,
  PEPValidationError,
  norm_printable,
  valid_dist_name,
  norm_dist_name,
  norm_dist_filename,
  join_dist_filename,
  norm_dist_version,
  norm_dist_author,
  norm_dist_classifier,
  norm_dist_keyword,
  norm_dist_url,
  norm_dist_extra,
  norm_dist_build,
  dist_build,
  norm_dist_compat,
  join_dist_compat,
  compress_dist_compat,
  norm_py_identifier,
  norm_entry_point_group,
  norm_entry_point_name,
  norm_entry_point_ref )

from .path import (
  PatternError,
  PathMatcher,
  PathFilter,
  partition,
  combine_ignore_patterns,
  contains )

from .template import (
  Template,
  Namespace,
  template_substitute,
  TemplateError,
  NamespaceError)

from .dist_file import (
  dist_base,
  dist_zip,
  dist_targz,
  dist_source_dummy,
  dist_source_targz,
  dist_binary_wheel,
  FileOutsideRootError,
  dist_iter,
  dist_copy )

from .pkginfo import (
  PkgInfoReq,
  PkgInfoAuthor,
  PkgInfoURL,
  PkgInfo )

from .builder import (
  Builder )

from .pyproj import (
  EntryPointError,
  PyProjBase )



from __future__ import annotations
import os
import os.path as osp
import sys
import sysconfig
import re
from copy import copy
import shutil
import subprocess
from logging import Logger
from pathlib import Path

from ..file import tail
from ..validate import (
  validating,
  ValidationError,
  ValidPathError,
  FileOutsideRootError )

from ..load_module import EntryPoint

from ..path import (
  subdir,
  resolve)

from ..template import (
  template_substitute,
  Namespace)
from ..pptoml import pyproj_targets

ERROR_REC = re.compile(r"error:", re.I)

#===============================================================================
class BuildCommandError(ValidationError):
  pass

#===============================================================================
class Builder:
  """Run build setup, compile, install commands

  Parameters
  ----------
  root:
    Path to root project directory
  targets:
  logger:

  """
  #-----------------------------------------------------------------------------
  def __init__(self,
    pyproj,
    root: str | Path,
    targets: pyproj_targets,
    logger: Logger):

    root = resolve(Path(root))

    self.pyproj = pyproj
    self.root = root
    # isolate (shallow) changes to targets
    self.targets = [copy(v) for v in targets]
    self.clean_dirs = [False]*len(self.targets)
    self.logger = logger
    self.namespace = Namespace({
      'root': root,
      'pptoml': pyproj.pptoml,
      'project': pyproj.project,
      'pyproj': pyproj.pyproj,
      'config_settings': pyproj.config_settings,
      'targets': targets,
      'env': os.environ,
      'config_vars': sysconfig.get_config_vars()},
      root=root)

  #-----------------------------------------------------------------------------
  def __enter__(self):
    return self

  #-----------------------------------------------------------------------------
  def __exit__(self, type, value, traceback):
    self.build_clean()

    # do not handle any exceptions here
    return False

  #-----------------------------------------------------------------------------
  def build_targets(self):
    for i, target in enumerate(self.targets):
      if not target.enabled:
        self.logger.info(f"Skipping targets[{i}], disabled for environment markers")
        continue

      # each target isolated (shallow) changes to namespace
      namespace = copy(self.namespace)

      # check paths
      for k in ('work_dir', 'src_dir', 'build_dir', 'prefix'):
        with validating(key = f"tool.pyproj.targets[{i}].{k}"):
          rel_path = target[k]
          rel_path = template_substitute(rel_path, namespace)

          if rel_path.is_absolute():
            abs_path = rel_path
          else:
            abs_path = self.root/rel_path

          abs_path = resolve(abs_path)

          if not subdir(self.root, abs_path, check=False):
            raise FileOutsideRootError(
              f"Must be within project root directory:"
              f"file = \"{abs_path}\",  root = \"{self.root}\"")

          if k in ('build_dir', 'prefix') and subdir(abs_path, self.root, check=False):
            raise ValidPathError(
              f"'{k}' cannot be project root directory:"
              f"file = \"{abs_path}\",  root = \"{self.root}\"")

          target[k] = abs_path
          namespace[k] = abs_path

      src_dir = target.src_dir
      build_dir = target.build_dir
      prefix = target.prefix
      work_dir = target.work_dir

      with validating(key = f"tool.pyproj.targets[{i}].src_dir"):
        if not src_dir.exists():
          raise ValidPathError(f"Source directory not found: {src_dir}")

        if not src_dir.is_dir():
          raise ValidPathError(f"Source directory not a directory: {src_dir}")

      with validating(key = f"tool.pyproj.targets[{i}]"):
        if subdir(build_dir, prefix, check=False):
          raise ValidPathError(
            f"'prefix' cannot be inside 'build_dir', which will be cleaned: {build_dir} > {prefix}")

      build_dirty = build_dir.exists() and any(build_dir.iterdir())

      if target.build_clean and build_dirty:
        raise ValidPathError(
          f"'build_dir' is not empty, please remove manually."
          f" If this was intended, set 'build_clean = false': {build_dir}")

      # create output directories
      for k in ['build_dir', 'prefix']:
        with validating(key = f"tool.pyproj.targets[{i}].{k}"):
          target[k].mkdir(parents=True, exist_ok=True)

      with validating(key = f"tool.pyproj.targets[{i}].options"):
        # original target options remain until evaluated
        options = target.options

        # top-level options updated in order of appearance
        _options = {}
        namespace['options'] = _options

        for k,v in options.items():
          v = template_substitute(v, namespace)
          # update target
          options[k] = v
          # update
          _options[k] = v

      with validating(key = f"tool.pyproj.targets[{i}].env"):
        # original target options remain until evaluated
        env = target.env

        # top-level options updated in order of appearance
        # copy of environment dict, each target isolated changes
        _env = copy(namespace['env'])
        namespace['env'] = _env

        for k,v in env.items():
          v = template_substitute(v, namespace)
          env[k] = v
          _env[k] = v

      for attr in ['setup_args', 'compile_args', 'install_args']:
        with validating(key = f"tool.pyproj.targets[{i}].{attr}"):
          value = target[attr]
          value = template_substitute(value, namespace)

          target[attr] = value
          namespace[attr] = value

      entry_point = EntryPoint(
        pyproj = self,
        root = self.root,
        name = f"tool.pyproj.targets[{i}]",
        logger = self.logger,
        entry = target.entry)

      log_dir = self.root/'build'/'logs'

      log_dir.mkdir(parents=True, exist_ok=True)

      runner = ProcessRunner(
        logger=self.logger,
        log_dir=log_dir,
        target_name=f"target_{i:02d}",
        env=_env)

      self.logger.info('\n'.join([
        f"targets[{i}]:",
        f"  work_dir: {work_dir}",
        f"  src_dir: {src_dir}",
        f"  build_dir: {build_dir}",
        f"  prefix: {prefix}",
        f"  log_dir: {log_dir}",
        "  options: " + ('\n' if target.options else 'none') + '\n'.join([
          f"    {k}: {v}" for k,v in target.options.items()]),
        "  env: " + ('\n' if target.env else 'default') + '\n'.join([
          f"    {k}: {v}" for k,v in target.env.items()])]))

      cwd = os.getcwd()

      # allow cleaning once the target is validated
      self.clean_dirs[i] = True

      try:
        os.chdir(work_dir)

        entry_point(
          options = target.options,
          work_dir = work_dir,
          src_dir = src_dir,
          build_dir = build_dir,
          prefix = prefix,
          setup_args = target.setup_args,
          compile_args = target.compile_args,
          install_args = target.install_args,
          build_clean = not build_dirty,
          runner = runner)

      finally:
        os.chdir(cwd)

  #-----------------------------------------------------------------------------
  def build_clean(self):
    for i, (target, clean) in enumerate(zip(self.targets, self.clean_dirs)):
      if not clean:
        continue

      build_dir = target.build_dir

      if build_dir is not None and build_dir.exists() and target.build_clean:
        self.logger.info(f"Removing build dir: {build_dir}")
        shutil.rmtree(build_dir)

#===============================================================================
class ProcessRunner:
  #-----------------------------------------------------------------------------
  def __init__(self,
      logger,
      log_dir: Path,
      target_name: str,
      env: dict):

    self.logger = logger
    self.log_dir = log_dir
    self.target_name = target_name
    self.commands = {}
    self.env = env

  #-----------------------------------------------------------------------------
  def run(self, args: list, env: dict = None):
    if len(args) == 0:
      raise ValueError(f"Command for {self.target_name} is empty.")

    cmd_exec = args[0]
    cmd_exec_src = shutil.which(cmd_exec)

    if cmd_exec_src is None:
      raise ValueError(
        f"Executable does not exist or has in-sufficient permissions: {cmd_exec}")

    cmd_exec_src = resolve(Path(cmd_exec_src))
    cmd_name = cmd_exec_src.name
    args = [str(cmd_exec_src)]+args[1:]

    cmd_hist = self.commands.setdefault(cmd_exec_src, [])
    cmd_idx = len(cmd_hist)
    cmd_hist.append(args)

    run_name = re.sub(r'[^\w]+', "_", cmd_name)
    run_id = f"{self.target_name}.{run_name}.{cmd_idx:02d}"

    stdout_file = self.log_dir/f"{run_id}.log"

    try:
      self.logger.info(f"Running {run_id!r}: "+' '.join(args))

      with open(stdout_file, 'wb') as fp:
        subprocess.run(
          args,
          shell=False,
          stdout=fp,
          stderr=subprocess.STDOUT,
          check=True,
          env=self.env)

    except subprocess.CalledProcessError as e:


      num_windows = 20
      window_size = 5
      with open(stdout_file, 'rb') as fp:
        lines = [
          (lineno,line)
          for lineno,line in enumerate(fp.read().decode('utf-8', errors='replace').splitlines())]

      suspect_linenos = [
        lineno
        for lineno,line in lines
        if ERROR_REC.search(line)]

      # suspect_linenos = suspect_linenos[:num_windows]

      extra = [
        '\n'.join(
          [f"{'':-<70}",f"{'':>4}⋮"]
          +[f"{j:>4d}| {line}" for j,line in lines[i:i+window_size]]
          +[f"{'':>4}⋮"])
        for i in suspect_linenos]

      m = len(lines)-num_windows

      if suspect_linenos:
        m = max(m, suspect_linenos[-1])

      last_lines = lines[m:]

      if last_lines:
        extra += [
          f"{'':-<70}",
          f"Last {len(last_lines)} lines of command output:",
          f"{'':>4}⋮"]

        extra += [
          f"{j:>4d}| {line}"
          for j,line in last_lines]

      extra += [
        f"{'END':>4}| [See log file: {stdout_file}]",
        f"{'':-<70}",]

      raise BuildCommandError(
        str(e),
        extra='\n'.join(extra)) from None
